/*
* JBoss, Home of Professional Open Source
* Copyright 2005, JBoss Inc., and individual contributors as indicated
* by the @authors tag. See the copyright.txt in the distribution for a
* full listing of individual contributors.
*
* This is free software; you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2.1 of
* the License, or (at your option) any later version.
*
* This software is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this software; if not, write to the Free
* Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
* 02110-1301 USA, or see the FSF site: http://www.fsf.org.
*/

package org.jboss.serial.studies;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.lang.reflect.Array;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.HashMap;
import java.util.Map;

import junit.framework.TestCase;

/**
 * This is a class with several studies I made about serialization's behaviors in order to input features into JBossSerialization
 *
 * Like, how to create object arrays
 *       how to access private fields
 *       etc...
 *
 * So, this class is not testing any specific functionality at JBossSerialization
 * $Id: StudyTest.java 137 2006-02-24 20:33:13Z csuconic $
 *
 * @author clebert suconic
 */
public class StudyTest extends TestCase
{

    static class TestChild implements Serializable
    {
        private transient int a=1;
        private transient int b=2;
        private transient Integer c=new Integer(3);
        private transient long d=4;

        public TestChild()
        {
            System.out.println("Constructor");
        }

        String test = "value";

        protected void hello()
        {
        }

    }


    public void testSimpleSerialization() throws Exception
    {
        ByteArrayOutputStream byteOut = new ByteArrayOutputStream();
        ObjectOutputStream outObj = new ObjectOutputStream(byteOut);

        TestChild filha = new TestChild();

        outObj.writeObject(filha);

        outObj.close();

        ByteArrayInputStream byteIn = new ByteArrayInputStream(byteOut
                .toByteArray());
        ObjectInputStream inp = new ObjectInputStream(byteIn);
        TestChild obj = (TestChild) inp.readObject();

        System.out.println("result = " + obj.test);

        assertTrue(filha != obj);
    }

    public void testForNameOnArray() throws Exception
    {
        Class clazz = Class.forName("[[Ljava.lang.String;");

        Object produced = Array.newInstance(clazz,0);

        System.out.println("produced = " + produced.getClass().getName());
    }

    public void testFindingDefaultConstructor() throws Exception
    {
        System.out.println("executing testFindingDefaultConstructor()");
        ByteArrayOutputStream byteOut = new ByteArrayOutputStream();
        ObjectOutputStream outObj = new ObjectOutputStream(byteOut);

        TestChild filha = new TestChild();

        outObj.writeObject(filha);

        outObj.close();

        ByteArrayInputStream byteIn = new ByteArrayInputStream(byteOut
                .toByteArray());
        ObjectInputStream inp = new ObjectInputStream(byteIn);
        TestChild obj = (TestChild) inp.readObject();
        assertTrue(filha != obj);

        System.out.println("result = " + obj.test);

        Constructor[] constructors = TestChild.class.getDeclaredConstructors();

        for (int i = 0; i < constructors.length; i++)
        {
            System.out.println("constructor[" + i + "]=" + constructors[i]);
        }

        Method[] methods = TestChild.class.getDeclaredMethods();
        for (int i = 0; i < methods.length; i++)
        {
            System.out.println("methods[" + i + "]=" + methods[i]);
        }

        Field[] fields = TestChild.class.getDeclaredFields();

        for (int i = 0; i < fields.length; i++)
        {
            System.out.println("fields[" + i + "]=" + fields[i]);
        }

        Constructor constructor = TestChild.class.getConstructor(new Class[]{});

        System.out.println("returned constructor=" + constructor);


    }

    public void testChangeFieldAccessor() throws Exception
    {
        System.out.println("executing testFindingDefaultConstructor()");

        Field[] fields = TestChild.class.getDeclaredFields();

        TestChild ob = new TestChild();
        for (int i = 0; i < fields.length; i++)
        {
            fields[i].setAccessible(true);
            System.out.println("isAccessible (first loop= " + i + ") = " + fields[i].isAccessible());

            if ((fields[i].getModifiers() & Modifier.TRANSIENT)!=0)
            {
                System.out.println(fields[i].getName() + " identified as transient");
            }


        }

        fields[0].setInt(ob,22);

        for (int i = 0; i < fields.length; i++)
        {
            System.out.println("Value of private field [ " + fields[i].getName() + "]= " + fields[i].get(ob));
        }


        Object newobj = TestChild.class.newInstance();

        fields[0].setInt(newobj,12345);

        for (int i = 0; i < fields.length; i++)
        {
            System.out.println("NewOBJ Value [ " + fields[i].getName() + "]= " + fields[i].get(newobj));
        }

        fields = TestChild.class.getDeclaredFields();
        for (int i = 0; i < fields.length; i++)
        {
            // The accessible should not change, it should be kept as false.
            System.out.println("isAccessible (second loop= " + i + ") = " + fields[i].isAccessible());
        }
    }

    public void testDatatypes() throws Exception
    {
        Field[] fields = TestChild.class.getDeclaredFields();

        for (int i=0;i<fields.length;i++)
        {
            if (fields[i].getType()==Integer.class)
            {
                System.out.println(fields[i].getName() + " is " + "Integer class");
            }

            if (fields[i].getType()==Integer.TYPE)
            {
                System.out.println(fields[i].getName() + " is " + "Integer primitive");
            }

        }
    }

    public void testReflectionInArraysLookingAtValues() throws Exception
    {
        String array1[][];
        array1 = new String[10][];
        for (int i=0;i<10;i++)
        {
            array1[i] = new String[i+1];
            for (int j=0;j<i+1;j++)
            {
                array1[i][j]=i + "-" + j;
            }
        }

        System.out.println("ArrayName -> " + array1.getClass().getName());

        exploreArray("",array1);

        exploreArray("",createDoubleArray());

        assertEquals(2,exploreDeepArray(createDoubleArray()));
    }

    /** How to navigate in a array using reflection */
    private void exploreArray(String message, Object array)
    {

        int size = Array.getLength(array);
        for (int i=0;i<size;i++)
        {
            Object obj = Array.get(array,i);
            if (obj.getClass().isArray())
            {
                exploreArray(message + " pos=" + i,obj);
            } else
            {
                System.out.println(message + " pos="+ i + " = " + obj);
            }
        }
    }


    /** How to discover the deep */
    private int exploreDeepArray(Object obj) throws Exception
    {
        int depth =0 ;

        Class currentClass = obj.getClass();

        while (currentClass.isArray())
        {
            depth++;
            currentClass = currentClass.getComponentType();
        }


        System.out.println("depth = " + depth);

        return depth;
    }

    /** How to create arrays using reflections */
    private Object createDoubleArray()
    {
        int x[] = new int[2];
        x[0]=10;
        x[1]=0;
        Object obj = Array.newInstance(String.class,x);

        for (int i=0;i<10;i++)
        {
            Object sub = Array.newInstance(String.class,i+1);
            for (int j=0;j<i+1;j++)
            {
                Array.set(sub,j,"SubValue " + i + " " + j);
            }
            Array.set(obj,i,sub);
        }

        return obj;
    }

    public void testIsInstanceOfLike()
    {
        // testing if HashMap is instanceof Map

        assertTrue(Map.class.isAssignableFrom(HashMap.class));
    }



    public void testArrayClass() throws Exception
    {
        try
        {
            Thread.currentThread().getContextClassLoader().loadClass("[Lorg.jboss.serial.studies.datatest.TestClass;");
            fail("This was supposed to fail");
        }
        catch (Exception e)
        {
        }
    }

}
